#!/usr/bin/env python3
"""
Powered by Meshtastic™ https://meshtastic.org/
"""

import time
from mmqtt.load_config import ConfigLoader
from mmqtt.mqtt_handler import get_mqtt_client
from mmqtt.argument_parser import handle_args, get_args

stay_connected = False

def main():
    _, args = get_args()
    config_file = args.config
    config = ConfigLoader.load_config_file(config_file)
    client = get_mqtt_client()
    handle_args() 
    
    if config.mode.listen == False:
        client.disconnect()
    else:
        while True:
            time.sleep(1)

if __name__ == "__main__":
    main()


from meshtastic import mqtt_pb2, portnums_pb2, mesh_pb2, telemetry_pb2
from meshtastic import protocols
from mmqtt.encryption import decrypt_packet
from mmqtt.load_config import ConfigLoader

def on_message(client, userdata, msg):
    config = ConfigLoader.get_config()

    if not config.mode.listen:
        return  # Do not process or print the message

    se = mqtt_pb2.ServiceEnvelope()
    try:
        se.ParseFromString(msg.payload)

        print ("")
        print ("Service Envelope:")
        print (se)

        mp = se.packet
    except Exception as e:
        print(f"*** ServiceEnvelope: {str(e)}")
        return
    
    if mp.HasField("encrypted") and not mp.HasField("decoded"):
        decoded_data = decrypt_packet(mp, config.channel.key)
        if decoded_data is not None:
            mp.decoded.CopyFrom(decoded_data)
        else:
            print("*** Failed to decrypt message — decoded_data is None")
            return

    print ("")
    print ("Message Packet:")
    print(mp)

    if mp.decoded.portnum == portnums_pb2.TEXT_MESSAGE_APP:
        try:
            from_str = getattr(mp, "from")
            from_id = '!' + hex(from_str)[2:]
            text_payload = mp.decoded.payload.decode("utf-8")

            print ("")
            print ("Message Payload:")
            print (mp.decoded)
            # print(f"{from_id}: {text_payload}")

        except Exception as e:
            print(f"*** TEXT_MESSAGE_APP: {str(e)}")
        
    elif mp.decoded.portnum == portnums_pb2.NODEINFO_APP:
        info = mesh_pb2.User()
        try:
            info.ParseFromString(mp.decoded.payload)

            print("")
            print("NodeInfo:")
            print(info)

        except Exception as e:
            print(f"*** NODEINFO_APP: {str(e)}")
        
    elif mp.decoded.portnum == portnums_pb2.POSITION_APP:
        pos = mesh_pb2.Position()
        try:
            pos.ParseFromString(mp.decoded.payload)

            print("")
            print("Position:")
            print(pos)

        except Exception as e:
            print(f"*** POSITION_APP: {str(e)}")

    elif mp.decoded.portnum == portnums_pb2.TELEMETRY_APP:
        telem = telemetry_pb2.Telemetry()
        try:
            # Parse the payload into the main telemetry message
            telem.ParseFromString(mp.decoded.payload)

            # Check and parse device_metrics if available
            if telem.HasField("device_metrics"):
                device_metrics = telem.device_metrics
                print("Device Metrics:")
                print(device_metrics)

            # Check and parse environment_metrics if available
            if telem.HasField("environment_metrics"):
                environment_metrics = telem.environment_metrics
                print("Environment Metrics:")
                print(environment_metrics)

            # Check and parse power_metrics if available
            if telem.HasField("power_metrics"):
                power_metrics = telem.power_metrics
                print("Power Metrics:")
                print(power_metrics)

        except Exception as e:
            print(f"*** TELEMETRY_APP: {str(e)}")

    else:
        # Attempt to process the decrypted or encrypted payload
        portNumInt = mp.decoded.portnum if mp.HasField("decoded") else None
        handler = protocols.get(portNumInt) if portNumInt else None

        pb = None
        if handler is not None and handler.protobufFactory is not None:
            pb = handler.protobufFactory()
            pb.ParseFromString(mp.decoded.payload)

        if pb:
            # Clean and update the payload
            pb_str = str(pb).replace('\n', ' ').replace('\r', ' ').strip()
            mp.decoded.payload = pb_str.encode("utf-8")
        print(mp)

        # rssi = getattr(mp, "rx_rssi")

        # # Device Metrics
        # device_metrics_dict = {
        #     'Battery Level': telem.device_metrics.battery_level,
        #     'Voltage': round(telem.device_metrics.voltage, 2),
        #     'Channel Utilization': round(telem.device_metrics.channel_utilization, 1),
        #     'Air Utilization': round(telem.device_metrics.air_util_tx, 1)
        # }

        # # Environment Metrics
        # environment_metrics_dict = {
        #     'Temp': round(telem.environment_metrics.temperature, 2),
        #     'Humidity': round(telem.environment_metrics.relative_humidity, 0),
        #     'Pressure': round(telem.environment_metrics.barometric_pressure, 2),
        #     'Gas Resistance': round(telem.environment_metrics.gas_resistance, 2)
        # }



import base64
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from meshtastic.protobuf import mesh_pb2


from mmqtt.utils import generate_hash

def decrypt_packet(mp, key):
    """Decrypt the encrypted message payload and return the decoded data."""
    try:
        key_bytes = base64.b64decode(key.encode('ascii'))

        # Build the nonce from message ID and sender
        nonce_packet_id = getattr(mp, "id").to_bytes(8, "little")
        nonce_from_node = getattr(mp, "from").to_bytes(8, "little")
        nonce = nonce_packet_id + nonce_from_node

        # Decrypt the encrypted payload
        cipher = Cipher(algorithms.AES(key_bytes), modes.CTR(nonce), backend=default_backend())
        decryptor = cipher.decryptor()
        decrypted_bytes = decryptor.update(getattr(mp, "encrypted")) + decryptor.finalize()

        # Parse the decrypted bytes into a Data object
        data = mesh_pb2.Data()
        data.ParseFromString(decrypted_bytes)
        return data

    except Exception as e:
        print(f"Failed to decrypt: {e}")
        return None

def encrypt_packet(channel, key, mp, encoded_message):
    """Encrypt a message."""

    try:
        mp.channel = generate_hash(channel, key)
        key_bytes = base64.b64decode(key.encode('ascii'))

        nonce_packet_id = getattr(mp, "id").to_bytes(8, "little")
        nonce_from_node = getattr(mp, "from").to_bytes(8, "little")
        
        # Put both parts into a single byte array.
        nonce = nonce_packet_id + nonce_from_node

        cipher = Cipher(algorithms.AES(key_bytes), modes.CTR(nonce), backend=default_backend())
        encryptor = cipher.encryptor()
        encrypted_bytes = encryptor.update(encoded_message.SerializeToString()) + encryptor.finalize()

        return encrypted_bytes
    
    except Exception as e:
        print(f"Failed to encrypt: {e}")
        return None


import argparse
import time

from mmqtt.load_config import ConfigLoader
from mmqtt.utils import validate_lat_lon_alt
from mmqtt.tx_message_handler import (
    send_position,
    send_text_message,
    send_nodeinfo,
    send_device_telemetry,
)

def get_args():
    """Define and parse command-line arguments."""
    parser = argparse.ArgumentParser(description="Meshtastic MQTT client")
    parser.add_argument('--config', type=str, default='config.json', help='Path to the config file')
    parser.add_argument('--message', type=str, help='The message to send')
    parser.add_argument('--nodeinfo', action='store_true', help='Send NodeInfo from my config')
    parser.add_argument('--telemetry', action='store_true', help='Send telemetry from my config')
    parser.add_argument('--lat', type=float, help='Latitude coordinate')
    parser.add_argument('--lon', type=float, help='Longitude coordinate')
    parser.add_argument('--alt', type=float, help='Altitude')
    parser.add_argument('--precision', type=int, help='Position Precision')
    parser.add_argument('--position', action='store_true', help='Send position from config unless overridden by --lat/lon/alt')
    parser.add_argument('--listen', action='store_true', help='Enable listening for incoming MQTT messages')

    args = parser.parse_args()
    return parser, args

def handle_args():
    parser, args = get_args()
    config = ConfigLoader.get_config()

    if args.message:
        send_text_message(args.message)
        print(f"Sending Message Packet to {config.message.destination_id}")
        time.sleep(3)
        return args

    if args.position:
        position = config.position

        lat = args.lat if args.lat is not None else position.lat
        lon = args.lon if args.lon is not None else position.lon
        alt = args.alt if args.alt is not None else position.alt
        precision = args.precision if args.precision is not None else position.precision

        validate_lat_lon_alt(parser, argparse.Namespace(lat=lat, lon=lon, alt=alt))

        send_position(lat, lon, alt, precision)
        print(f"Sending Position Packet to {config.message.destination_id}")
        time.sleep(3)
        return args

    if args.nodeinfo:
        node = config.nodeinfo
        send_nodeinfo(node.short_name, node.long_name, node.hw_model)
        print(
            "Sending NodeInfo:\n"
            f"  Short Name: {node.short_name}\n"
            f"  Long Name:  {node.long_name}\n"
            f"  Hardware Model: {node.hw_model}"
        )
        time.sleep(3)

    if args.telemetry:
        telemetry = config.telemetry
        send_device_telemetry(
            battery_level=telemetry.battery_level,
            voltage=telemetry.voltage,
            chutil=telemetry.chutil,
            airtxutil=telemetry.airtxutil,
            uptime=telemetry.uptime
        )
        print(
            "Sending Telemetry:\n"
            f"  Battery Level:        {telemetry.battery_level}%\n"
            f"  Voltage:              {telemetry.voltage}V\n"
            f"  Channel Utilization:  {telemetry.chutil}%\n"
            f"  Air Tx Utilization:   {telemetry.airtxutil}%\n"
            f"  Uptime:               {telemetry.uptime}s"
        )
        time.sleep(3)


    if args.listen:
        from mmqtt.mqtt_handler import get_mqtt_client
        from mmqtt.rx_message_handler import on_message

        # Load config and override mode.listen from CLI
        config = ConfigLoader.get_config()
        config.mode.listen = True

        print("Starting MQTT listener (press Ctrl+C to stop)...")

        # Start the MQTT client and set up the message handler
        client = get_mqtt_client()
        client.on_message = on_message
        client.loop_start()

    return args


import time
import re
import random

from meshtastic import portnums_pb2, mesh_pb2, mqtt_pb2, telemetry_pb2
from mmqtt.utils import generate_hash, get_message_id
from mmqtt.encryption import encrypt_packet
from mmqtt.load_config import ConfigLoader
from mmqtt.mqtt_handler import get_mqtt_client

_config = None

def _get_config():
    global _config
    if _config is None:
        _config = ConfigLoader.get_config()
    return _config

message_id = random.getrandbits(32)

def publish_message(payload_function, portnum, **kwargs):
    """Send a message of any type."""
    try:
        config = _get_config()
        client = get_mqtt_client()
        payload = payload_function(portnum=portnum, **kwargs)
        topic = f"{config.mqtt.root_topic}{config.channel.preset}/{config.nodeinfo.id}"
        client.publish(topic, payload)
    except Exception as e:
        print(f"Error while sending message: {e}")

def create_payload(data, portnum, want_response=False, bitfield=1):
    """Generalized function to create a payload."""
    encoded_message = mesh_pb2.Data()
    encoded_message.portnum = portnum
    encoded_message.payload = data.SerializeToString() if hasattr(data, "SerializeToString") else data
    encoded_message.want_response = want_response
    encoded_message.bitfield = bitfield
    return generate_mesh_packet(encoded_message)

def generate_mesh_packet(encoded_message):
    """Generate the final mesh packet."""
    config = _get_config()
    global message_id
    message_id = get_message_id(message_id)
    
    mesh_packet = mesh_pb2.MeshPacket()
    mesh_packet.id = message_id
    setattr(mesh_packet, "from", config.nodeinfo.number)
    mesh_packet.to = int(config.message.destination_id)
    mesh_packet.want_ack = False
    mesh_packet.channel = generate_hash(config.channel.preset, config.channel.key)
    mesh_packet.hop_limit = 3
    mesh_packet.hop_start = 3

    if config.channel.key == "":
        mesh_packet.decoded.CopyFrom(encoded_message)
    else:
        mesh_packet.encrypted = encrypt_packet(config.channel.preset, config.channel.key, mesh_packet, encoded_message)

    service_envelope = mqtt_pb2.ServiceEnvelope()
    service_envelope.packet.CopyFrom(mesh_packet)
    service_envelope.channel_id = config.channel.preset
    service_envelope.gateway_id = config.nodeinfo.id

    return service_envelope.SerializeToString()

########## Specific Message Handlers ##########

def send_text_message(message):
    """Send a text message."""
    def create_text_payload(portnum, message_text):
        data = message_text.encode("utf-8")
        return create_payload(data, portnum)
    
    publish_message(create_text_payload, portnums_pb2.TEXT_MESSAGE_APP, message_text=message)

def send_nodeinfo(long_name, short_name, hw_model):
    """Send node information."""
    def create_nodeinfo_payload(portnum, node_long_name, node_short_name, node_hw_model):
        config = _get_config()
        data = mesh_pb2.User(
            id=config.nodeinfo.id,
            long_name=node_long_name,
            short_name=node_short_name,
            hw_model=node_hw_model
        )
        return create_payload(data, portnum)
    
    publish_message(create_nodeinfo_payload, portnums_pb2.NODEINFO_APP, 
                 node_long_name=long_name, node_short_name=short_name, node_hw_model=hw_model)

def send_position(lat, lon, alt, pre):
    """Send position details."""
    def create_position_payload(portnum, lat, lon, alt, pre):
        pos_time = int(time.time())
        latitude = int(float(lat) * 1e7)
        longitude = int(float(lon) * 1e7)
        altitude_units = 1 / 3.28084 if 'ft' in str(alt) else 1.0
        altitude = int(altitude_units * float(re.sub('[^0-9.]', '', str(alt))))

        data = mesh_pb2.Position(
            latitude_i=latitude,
            longitude_i=longitude,
            altitude=altitude,
            time=pos_time,
            location_source="LOC_MANUAL",
            precision_bits=pre
        )
        return create_payload(data, portnum)

    publish_message(create_position_payload, portnums_pb2.POSITION_APP, lat=lat, lon=lon, alt=alt, pre=pre)

def send_device_telemetry(battery_level, voltage, chutil, airtxutil, uptime):
    """Send telemetry data."""
    def create_telemetry_payload(portnum, battery_level, voltage, chutil, airtxutil, uptime):
        data = telemetry_pb2.Telemetry(
            time=int(time.time()),
            device_metrics=telemetry_pb2.DeviceMetrics(
                battery_level=battery_level,
                voltage=voltage,
                channel_utilization=chutil,
                air_util_tx=airtxutil,
                uptime_seconds=uptime
            )
        )
        return create_payload(data, portnum)

    publish_message(create_telemetry_payload, portnums_pb2.TELEMETRY_APP, 
                 battery_level=battery_level, voltage=voltage, chutil=chutil, airtxutil=airtxutil, uptime=uptime)


import time
import ssl
import paho.mqtt.client as mqtt
from mmqtt.rx_message_handler import on_message
from mmqtt.load_config import ConfigLoader

_client_instance = None

auto_reconnect = True
auto_reconnect_delay = 1

def set_topic():
    config = ConfigLoader.get_config()
    print(f"set_topic: {config.mqtt.root_topic}{config.channel.preset}/")
    node_name = '!' + hex(config.nodeinfo.number)[2:]
    subscribe_topic = config.mqtt.root_topic + config.channel.preset + "/#"
    publish_topic = config.mqtt.root_topic + config.channel.preset + "/" + node_name
    return subscribe_topic, publish_topic

def get_mqtt_client():
    """Get or create the MQTT client instance."""
    global _client_instance

    if _client_instance is None:
        _client_instance = connect_mqtt()

    return _client_instance


def connect_mqtt():
    """Create and connect the MQTT client."""
    config = ConfigLoader.get_config()
    client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, client_id="", clean_session=True, userdata=None)
    client.on_connect = on_connect
    client.on_disconnect = on_disconnect
    client.on_message = on_message

    if "tls_configured" not in connect_mqtt.__dict__:
        connect_mqtt.tls_configured = False

    if not client.is_connected():
        try:
            if ':' in config.mqtt.broker:
                config.mqtt.broker, config.mqtt.port = config.mqtt.broker.split(':')
                config.mqtt.port = int(config.mqtt.port)
            client.username_pw_set(config.mqtt.user, config.mqtt.password)
            if config.mqtt.port == 8883 and not connect_mqtt.tls_configured:
                client.tls_set(ca_certs="cacert.pem", tls_version=ssl.PROTOCOL_TLSv1_2)
                client.tls_insecure_set(False)
                connect_mqtt.tls_configured = True
            client.connect(config.mqtt.broker, config.mqtt.port, 60)
        except Exception as e:
            print(e)

        client.loop_start()
        time.sleep(1)
    return client


def on_disconnect(client, userdata, flags, reason_code, properties=None):
    print("client is disconnected")
    if reason_code != 0:
        if auto_reconnect:
            print("attempting to reconnect in " + str(auto_reconnect_delay) + " second(s)")
            time.sleep(auto_reconnect_delay)
            connect_mqtt()


def on_connect(client, userdata, flags, reason_code, properties=None):
    if client.is_connected():
        print("client is connected")

    if reason_code == 0:
        subscribe_topic, publish_topic = set_topic()
        print(f"Publish Topic is: {publish_topic}")
        print(f"Subscribe Topic is: {subscribe_topic}")
        client.subscribe(subscribe_topic)
    else:
        print("Failed to connect, return code %d\n", reason_code)


import os
import json
from types import SimpleNamespace
from meshtastic import BROADCAST_NUM

class ConfigLoader:
    _config = None

    @staticmethod
    def load_config_file(filename):
        if ConfigLoader._config is not None:
            return ConfigLoader._config  # Return already loaded config

        script_dir = os.path.dirname(os.path.abspath(__file__))
        config_path = os.path.join(script_dir, filename)

        # Fallback to config-example.json if config.json is missing
        if not os.path.exists(config_path):
            fallback_path = os.path.join(script_dir, "config-example.json")
            if os.path.exists(fallback_path):
                print(f"{filename} not found. Falling back to config-example.json.")
                config_path = fallback_path
            else:
                raise FileNotFoundError(f"Neither {filename} nor config-example.json found.")

        with open(config_path, 'r') as config_file:
            conf = json.load(config_file)

        # Expand default key
        conf["channel"]["key"] = "1PG7OiApB1nwvP+rz05pAQ==" if conf["channel"]["key"] == "AQ==" else conf["channel"]["key"]

        # Create keys not in config
        conf["nodeinfo"]["number"] = int(conf["nodeinfo"]["id"].replace("!", ""), 16)
        
        # Convert "False"/"True" strings to bool
        conf["mode"]["listen"] = conf["mode"]["listen"].lower() == "true"

        # Convert to nested SimpleNamespace
        def dict_to_namespace(data):
            if isinstance(data, dict):
                return SimpleNamespace(**{k: dict_to_namespace(v) for k, v in data.items()})
            return data
        
        ConfigLoader._config = dict_to_namespace(conf)
        return ConfigLoader._config

    @staticmethod
    def get_config(path="config.json"):
        if ConfigLoader._config is None:
            ConfigLoader.load_config_file(path)
        return ConfigLoader._config


if __name__ == "__main__":
    config = ConfigLoader.load_config_file('config.json')
    print(json.dumps(config, default=lambda o: o.__dict__, indent=4))


from meshtastic.protobuf import portnums_pb2
import base64
import random

def get_portnum_name(portnum) -> str:
    """For Logging: Retrieve the name of the port number from the protobuf enum."""
    try:
        return portnums_pb2.PortNum.Name(portnum)  # Use protobuf's enum name lookup
    except ValueError:
        return f"Unknown ({portnum})"  # Handle unknown port numbers gracefully
    
def protobuf_to_clean_string(proto_message) -> str:
    """For Logging: Convert protobuf message to string and remove newlines."""
    return str(proto_message).replace('\n', ' ').replace('\r', ' ').strip()

def xor_hash(data: bytes) -> int:
    """Compute an XOR hash from bytes."""
    result = 0
    for char in data:
        result ^= char
    return result

def generate_hash(name: str, key: str) -> int:
    """generate the channel number by hashing the channel name and psk"""
    replaced_key = key.replace('-', '+').replace('_', '/')
    key_bytes = base64.b64decode(replaced_key.encode('utf-8'))
    h_name = xor_hash(bytes(name, 'utf-8'))
    h_key = xor_hash(key_bytes)
    result: int = h_name ^ h_key
    return result

def get_message_id(rolling_message_id: int, max_message_id: int = 4294967295) -> int:
    """Increment the message ID with sequential wrapping and add a random upper bit component to prevent predictability."""
    rolling_message_id = (rolling_message_id + 1) % (max_message_id & 0x3FF + 1)
    random_bits = random.randint(0, (1 << 22) - 1) << 10
    message_id = rolling_message_id | random_bits
    return message_id


def validate_lat_lon_alt(parser, args) -> None:
    # Check if --alt is provided
    if args.alt:
        if not args.lat or not args.lon:
            parser.error('--alt should not be provided without --lat and --lon.')

    # Check if lat and lon are provided
    if args.lat or args.lon or args.pre:
        # If one of lat lon or pre is provided, ensure lat and lon are provided
        if not (args.lat and args.lon):
            parser.error('If you specify --lat, --lon, or --pre, you must specify both --lat and --lon.')

